@@ -31,20 +31,48 @@ struct sqsh_super_block {
 static int probe_squashfs(blkid_probe pr, const struct blkid_idmag *mag)
 {
 	struct sqsh_super_block *sq;
+	uint16_t vermaj;
+	uint16_t vermin;
 
 	sq = blkid_probe_get_sb(pr, mag, struct sqsh_super_block);
 	if (!sq)
-		return -1;
+		return errno ? -errno : 1;
+
+	vermaj = le16_to_cpu(sq->s_major);
+	vermin = le16_to_cpu(sq->s_minor);
+	if (vermaj < 4)
+		return 1;
+
+	blkid_probe_sprintf_version(pr, "%u.%u", vermaj, vermin);
+
+	return 0;
+}
+
+static int probe_squashfs3(blkid_probe pr, const struct blkid_idmag *mag)
+{
+	struct sqsh_super_block *sq;
+	uint16_t vermaj;
+	uint16_t vermin;
+
+	sq = blkid_probe_get_sb(pr, mag, struct sqsh_super_block);
+	if (!sq)
+		return errno ? -errno : 1;
+
+	if (strcmp(mag->magic, "sqsh") == 0) {
+		vermaj = be16_to_cpu(sq->s_major);
+		vermin = be16_to_cpu(sq->s_minor);
+	} else {
+		vermaj = le16_to_cpu(sq->s_major);
+		vermin = le16_to_cpu(sq->s_minor);
+	}
+
+	if (vermaj > 3)
+		return 1;
+
+	blkid_probe_sprintf_version(pr, "%u.%u", vermaj, vermin);
+
+	blkid_probe_set_block_size(pr, 1024);
 
-	if (strcmp(mag->magic, "sqsh") == 0 ||
-	    strcmp(mag->magic, "qshs") == 0)
-		blkid_probe_sprintf_version(pr, "%u.%u",
-				sq->s_major,
-				sq->s_minor);
-	else
-		blkid_probe_sprintf_version(pr, "%u.%u",
-				swab16(sq->s_major),
-				swab16(sq->s_minor));
 	return 0;
 }
 
@@ -55,14 +83,21 @@ const struct blkid_idinfo squashfs_idinfo =
 	.probefunc	= probe_squashfs,
 	.magics		=
 	{
-		{ .magic = "sqsh", .len = 4 },
-		{ .magic = "hsqs", .len = 4 }, /* swap */
-
-		/* LZMA version */
-		{ .magic = "qshs", .len = 4 },
-		{ .magic = "shsq", .len = 4 }, /* swap */
+		{ .magic = "hsqs", .len = 4 },
 		{ NULL }
 	}
 };
 
+const struct blkid_idinfo squashfs3_idinfo =
+{
+	.name		= "squashfs3",
+	.usage		= BLKID_USAGE_FILESYSTEM,
+	.probefunc	= probe_squashfs3,
+	.magics		=
+	{
+		{ .magic = "sqsh", .len = 4 }, /* big endian */
+		{ .magic = "hsqs", .len = 4 }, /* little endian */
+		{ NULL }
+	}
+};
 
